import os
from cement.utils import fs
from cement import App, TestApp, init_defaults
from cement.core.exc import CaughtSignal
import sqlite3

from benchgrape.config import DB_FILE
from benchgrape.controllers.benchmark import Benchmark
from benchgrape.controllers.test_data import TestData
from benchgrape.controllers.websocket_stability import Stability
from benchgrape.core.db import Mapper
from .core.exc import BenchGrapeError
from .controllers.base import Base

# configuration defaults
CONFIG = init_defaults('benchgrape')
CONFIG['benchgrape']['db_file'] = DB_FILE


def extend_sqlite(app):
    app.log.info('initializing sqlite')
    db_file = app.config.get('benchgrape', 'db_file')

    # ensure that we expand the full path
    db_file = fs.abspath(db_file)
    app.log.info('sqlite database file is: %s' % db_file)

    # ensure our parent directory exists
    db_dir = os.path.dirname(db_file)
    if not os.path.exists(db_dir):
        os.makedirs(db_dir)
    db_conn = sqlite3.connect(db_file, isolation_level='EXCLUSIVE')
    app.extend('db', db_conn)


def migrate(app):
    for klass in inheritors(Mapper):
        klass(app.db, app.log).init_db()


class BenchGrape(App):
    """Bench Grape primary application."""

    class Meta:
        label = 'benchgrape'

        # configuration defaults
        config_defaults = CONFIG

        # call sys.exit() on close
        close_on_exit = True

        # load additional framework extensions
        extensions = [
            'yaml',
            'colorlog',
            'jinja2',
        ]

        # configuration handler
        config_handler = 'yaml'

        # configuration file suffix
        config_file_suffix = '.yml'

        # set the log handler
        log_handler = 'colorlog'

        # set the output handler
        output_handler = 'jinja2'

        # register handlers
        handlers = [
            Base,
            TestData,
            Stability,
            Benchmark,
        ]

        hooks = [
            ('post_setup', extend_sqlite),
            ('post_setup', migrate),
        ]


class BenchGrapeTest(TestApp,BenchGrape):
    """A sub-class of BenchGrape that is better suited for testing."""

    class Meta:
        label = 'benchgrape'


def main():
    with BenchGrape() as app:
        try:
            app.run()

        except AssertionError as e:
            print('AssertionError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except BenchGrapeError as e:
            print('BenchGrapeError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)
            print('\n%s' % e)
            app.exit_code = 0


def inheritors(klass):
    subclasses = set()
    work = [klass]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return subclasses


if __name__ == '__main__':
    main()



import json
import logging
from requests.auth import HTTPBasicAuth

logger = logging.getLogger(__name__)


def login(client, user, password):
    if not user:
        logger.info('no user set in env var, falling back to admin')
        user = 'admin'

    if not password:
        logger.info('no password set in env var, falling back to admin')
        password = 'admin'

    with client.get(
            url="/api/accounts/session/",
            auth=HTTPBasicAuth(user, password),
            catch_response=True,
            verify=False,
    ) as response:
        print('login response: %i' % response.status_code)
        content = json.loads(response.content) if response.content else {}

        if response.status_code != 200 or 'detail' in content:
            msg = 'login failed %i %s' % (
                    response.status_code, content.get('detail')
            )
            print(msg)
            response.failure(msg)
            return
        return content['authtoken']


class PreparedStatementsMixin(object):
    """
    prepare statements for common requests like get_users or get_history,
    to be continued, those shall use the self.send in an abstract way so this
    mixin can be used by a HttpLocust OR a WampLocust Taskset equally.
    """
    org = None

    def send(self, ns, action, *args, **kwargs):
        raise NotImplementedError(
            'this mixin requires to be used with a taskset or adapter '
            'implementing a "send" method. see WampTaskSet implementation.'
        )

    @staticmethod
    def get_parameters(**kwargs):
        params = {}
        description = ''

        for param, value in kwargs.items():
            if value in ["", None]:
                description += f'without {param} '
            else:
                description += f'with {param} "{value}" '
                params[param] = value

        params['description'] = description

        return params

    def utils__connect_stack(self):
        self.organizations__get_organization()
        self.organizations__join()
        self.channels__get_overview()

    def utils__disconnect_stack(self):
        self.organizations__leave()

    def organizations__join(self):
        self.send('organizations', 'join', str(self.org))

    def organizations__leave(self):
        self.send('organizations', 'leave', str(self.org))

    def users__get_users(self, membership=None, query=None):
        params = self.get_parameters(membership=membership, query=query)

        self.send('users', 'get_users', str(self.org), **params)

    def rooms__get_rooms(self, membership=None, query=None):
        params = self.get_parameters(membership=membership, query=query)

        self.send('rooms', 'get_rooms', str(self.org), **params)

    def channels__get_history(self, ch, params=None):
        params = params or {}
        self.send('channels', 'get_history', ch, **params)

    def search__search_channels(self, search_text, limit=25, current_channel=None):
        self.send(
            'search', 'search_channels',
            self.org, search_text, limit, current_channel
        )

    def search__search_users(self, search_text, limit=25, current_channel=None):
        self.send(
            'search', 'search_users',
            self.org, search_text, limit, current_channel
        )

    def channels__get_overview(self, params=None):
        self.send(
            'channels', 'get_overview',
            str(self.org), params
        )

    def channels__post(self, ch_id, message):
        self.send(
            'channels', 'post',
            ch_id, '"' + str(message) + '"'
        )

    def channels__join(self, ch_id):
        self.send(
            'channels', 'join',
            ch_id
        )

    def channels__leave(self, ch_id):
        self.send(
            'channels', 'leave',
            ch_id
        )

    def channels__read(self, ch_id, msg_id):
        self.send(
            'channels', 'read',
            ch_id, msg_id
        )

    def channels__set_typing(self, ch_id, typing):
        self.send(
            'channels', 'set_typing',
            ch_id, typing
        )

    def organizations__get_organization(self):
        self.send(
            'organizations', 'get_organization',
            str(self.org), return_channels=False, return_user=False
        )



import sqlite3


class Mapper(object):
    def __init__(self, db_conn, logger):
        self.db = db_conn
        self.logger = logger

    def init_db(self):
        raise NotImplemented('mappers must implement init_db.')


class RuntimeConfigMapper(Mapper):
    db = None
    tables = ['runtime_config']

    def init_db(self):
        c = self.db.cursor()
        # Create tables
        self.logger.info('init runtime config table...')
        c.execute(
            '''CREATE TABLE IF NOT EXISTS runtime_config (
                recycle_mode boolean
            )'''
        )

    def enable_user_recycliing(self):
        c = self.db.cursor()
        c.execute("UPDATE runtime_config SET recycle_mode='TRUE'")
        self.db.commit()
        self.logger.info('enabled user recycle mode.')

    def disable_user_recycling(self):
        c = self.db.cursor()
        c.execute("UPDATE runtime_config SET recycle_mode='FALSE'")
        self.db.commit()
        self.logger.info('disabled user recycle mode.')


class TestDataMapper(Mapper):
    db = None
    tables = ['organizations', 'users', 'groups', 'private_conversations']

    def drop_db(self):
        c = self.db.cursor()
        for tb in self.tables:
            self.logger.info('dropping table %s' % tb)
            c.execute("DROP TABLE %s" % tb)
        self.logger.info('committing...')
        self.db.commit()
        self.logger.info('all gone!')

    def init_db(self):
        c = self.db.cursor()
        # Create tables
        try:
            self.logger.info('init organizations table...')
            c.execute(
                '''CREATE TABLE IF NOT EXISTS organizations (
                id INTEGER PRIMARY KEY, url text, name text, subdomain text,
                CONSTRAINT unique_id UNIQUE (id))'''
            )
            self.logger.info('init user table...')
            c.execute(
                '''CREATE TABLE IF NOT EXISTS users (
                id INTEGER PRIMARY KEY, username text, password text,
                token text, connected boolean,
                CONSTRAINT unique_id UNIQUE (id))'''
            )
            self.logger.info('init groups table...')
            c.execute('''CREATE TABLE IF NOT EXISTS groups (
                        id INTEGER PRIMARY KEY, name text,
                        CONSTRAINT unique_id UNIQUE (id)
                    )''')
            self.logger.info('init PM table...')
            c.execute('''CREATE TABLE IF NOT EXISTS private_conversations (
                        id INTEGER PRIMARY KEY, creator text,
                        CONSTRAINT unique_id UNIQUE (id)
                    )''')
            self.db.commit()
        except sqlite3.OperationalError as ex:
            self.logger.error(ex)
            self.db.rollback()

        self.logger.info('nice, so many tables!')

    def stats(self):
        cur = self.db.cursor()
        stats = {}
        for tb in self.tables:
            cur.execute("select count(*) from %s" % tb)
            stats[tb] = cur.fetchone()[0]

        return stats

    def sync_db(self, organization, users=None, groups=None, private_conversations=None):
        c = self.db.cursor()
        try:
            self.logger.info('creating %s organization...' % organization['name'])
            c.execute(
                "INSERT OR REPLACE INTO organizations (id, url, name, subdomain) "
                "VALUES (%i, '%s', '%s', '%s')" % (
                    int(organization['id']), organization['url'],
                    organization['name'],  organization['subdomain']
                ))

            self.logger.info('inserting %s users...' % len(users))
            for u in users or []:
                c.execute(
                    "INSERT OR REPLACE INTO users (id, username, password, token, connected) "
                    "VALUES (%i, '%s', '%s', '%s', 'FALSE')" % (
                        int(u['id']), u['username'], u['password'], u['token']
                    ))

            self.logger.info('inserting %s groups...' % len(groups))
            for g in groups or []:
                c.execute(
                    "INSERT OR REPLACE INTO groups (id, name) "
                    "VALUES (%i, '%s')" % (int(g['id']), g['name'])
                )

            self.logger.info('inserting %s private conversations...' % len(
                private_conversations))
            for pc in private_conversations or []:
                c.execute(
                    "INSERT OR REPLACE INTO private_conversations (id, creator) "
                    "VALUES (%i, '%s')" % (int(pc['id']), pc['creator'])
                )
            self.logger.info('benchmark data in sync with config file.')
            self.logger.info(
                'data heap now consists of %s' % (
                    ["%s: %s" % (k, v) for k, v in self.stats().items()]
                ))

            self.db.commit()
        except sqlite3.OperationalError as ex:
            self.logger.error(ex)
            self.db.rollback()
            self.logger.error('rolled back to initial state!')

    def _pick_unique_user(self):
        self.logger.info('picking a user')
        c = self.db.cursor()
        user = c.execute("SELECT id, username, password, token "
                         "FROM users "
                         "WHERE connected='FALSE' "
                         "ORDER BY RANDOM() "
                         "LIMIT 1 "
                         ).fetchone()

        if not user:
            self.db.rollback()
            raise RuntimeError('no users left to use.')
        upt = c.execute("UPDATE users SET connected='TRUE' WHERE id=%s" % user[0])
        assert upt.rowcount == 1
        self.db.commit()
        self.logger.info('picked user: %s' % user[1])
        return {
            'id': user[0], 'username': user[1],
            'password': user[2], 'token': user[3]
        }

    def _pick_user(self):
        self.logger.info('picking a user')
        c = self.db.cursor()
        user = c.execute("SELECT id, username, password, token "
                         "FROM users "
                         "ORDER BY RANDOM() "
                         "LIMIT 1 "
                         ).fetchone()

        if not user:
            self.db.rollback()
            raise RuntimeError('no users left to use.')

        self.logger.info('picked user: %s' % user[1])
        return {
            'id': user[0], 'username': user[1],
            'password': user[2], 'token': user[3]
        }

    def pick_user(self):
        c = self.db.cursor()
        unique_mode = c.execute(
            "SELECT recycle_mode FROM runtime_config  LIMIT 1 "
        ).fetchone()

        user = None
        if not unique_mode:
            return self._pick_user()

        while user is None:
            # try:
            user = self._pick_unique_user()
            # except Exception as ex:
            #     self.logger.error(ex)
            #     user = self.pick_user()
        return user

    def relieve_user(self, user_id):
        c = self.db.cursor()
        c.execute("UPDATE users SET connected='FALSE' WHERE id=%s" % user_id)
        self.db.commit()
        self.logger.info('user %s relieved' % user_id)

    def get_host(self):
        c = self.db.cursor()
        c.execute("SELECT url from organizations")
        return c.fetchone()[0]

    def get_organization(self):
        c = self.db.cursor()
        c.execute("SELECT * from organizations")
        org = c.fetchone()
        return {
            'id': org[0],
            'host': org[1],
            'name': org[2],
            'subdomain': org[3],
        }

    def sanitize(self):
        c = self.db.cursor()
        c.execute("UPDATE users SET connected='FALSE'")
        self.db.commit()
        self.logger.info('all users relieved')



import json
import logging
from json import JSONDecodeError
from random import choice
from string import digits, ascii_letters
import gevent
import six
import uuid
import ssl
from locust import TaskSet
from locust.events import request_success, request_failure, EventHook
from time import time
from urllib.parse import urlparse

import sqlite3
from websocket import create_connection, WebSocketConnectionClosedException


# todo add asserts to statements to validate response.
# todo dockerize all locust things and swarm it with docker compose / scale
from benchgrape.config import DB_FILE, WS
from benchgrape.core.db import TestDataMapper
from benchgrape.core.statement import login
from cement.utils import fs

logger = logging.getLogger(__name__)

db_file = fs.abspath(DB_FILE)
db_conn = sqlite3.connect(db_file, isolation_level='EXCLUSIVE')


def generate_call_id():
    return ''.join(choice(digits + ascii_letters) for _ in range(6))


websocket_dropped = EventHook()


def on_websocket_dropped(ex, *args, **kwargs):
    logger.exception(
        'could not receive data from websocket, '
        'reconnecting... exception was: %s', ex
    )


websocket_dropped += on_websocket_dropped


class MembershipSession(object):
    """
    represents a util class for a user session. can query channels a user
    has access to and save them, etc..
    """
    _ch_mbs = None
    user = None

    def __init__(self, task_set, user, organizatoin):
        self.user = user
        self.client = task_set.client
        self.wamp_client = task_set
        self.host = task_set.host
        self.auth_token = login(self.client, self.user['username'], self.user['password'])
        self.org = organizatoin['id']

    def get_client(self, task_set):
        if not hasattr(task_set, 'client'):
            return self.get_client(task_set.parent)
        return task_set.client

    # @property
    # def channel_memberships(self):
    #     if not self._ch_mbs:
    #         self.ch_mbs = self.wamp_client.send(
    #             'rooms', 'get_rooms', str(self.user['org_id']),
    #             **{'membership': True, 'page_size': 1000}
    #         )
    #     return self._ch_mbs

    @property
    def channel_memberships(self):
        # todo save to db - only when missing fetch all of them.
        if not self._ch_mbs:
            params = {
                "args": [self.user['org_id'], {'membership': True, 'page_size': 1000}],
                "action": 'get_rooms',
                "ns": 'rooms',
            }

            with self.client.post(
                    "/lp/rpc/", data=json.dumps(params),
                    catch_response=True,
                    headers={
                        "Authorization": "Token {}".format(self.auth_token),
                        "Content-Type": "application/json; charset=utf-8",
                    },
                    verify=False,
            ) as response:
                logger.debug('get_rooms response: %i' % response.status_code)
                if response.status_code < 200 or response.status_code > 299:
                    response.failure(response.status_code)
                else:
                    self._ch_mbs = json.loads(
                        response.content
                    )['response']['results'] if response.content else {}

        return self._ch_mbs

    def get_random_channel(self):
        return choice(self.channel_memberships)['id']


class PubSubMixin(object):
    """
    mixin to subscribe to the org right away when the set starts.
    """
    def on_start(self):
        logger.info("on_start called")
        super(PubSubMixin, self).on_start()
        self.organizations__join()


class PickleUserMixin(object):
    mapper = None
    organization = None

    def __init__(self, *args, **kwargs):
        super(PickleUserMixin, self).__init__(*args, **kwargs)
        self.mapper = mapper = TestDataMapper(db_conn, logger)
        self.host = mapper.get_host()
        self.user = mapper.pick_user()
        self.org = mapper.get_organization()

    def on_quit(self, *args, **kwargs):
        super(PickleUserMixin, self).on_quit(*args, **kwargs)
        self.mapper.relieve_user(self.user['id'])


class WampTaskSet(PickleUserMixin, TaskSet):
    """
    task set specially for wamp driven connections (v1 of course :D).
    """

    """
    pending requests, means to store sent request ids and meta info like
    request time and call type / name.
    those infos shall be deleted when the response is received
    asynchronously via _receive.
    """
    pending = None
    ws = None
    """
    user="thread" id aka random number. not a chatgrape user id
    """
    user_id = None
    """HistoryTaskSet
    store the auth token received from the login method (http "rest").
    """
    auth_token = None

    """
    falls back to ws://localhost:8000/ws
    """
    ws_location = None
    """
    ws or wss
    """
    ws_prot = None
    """
    I could tell you a joke here but it would be a waste of time.
    this was aleady a little waste of time.
    """
    org = None

    """
    we dont know the id of the first call so we dont find a start_time.
    the first call though is the create_connection handshake so we try
    to remember if we just opened the connection so we can ignore the first
    recv event.
    """
    awaiting_login_confirmation = False

    _mb = None

    @property
    def membership(self):
        if not self._mb:
            self._mb = MembershipSession(task_set=self, user=self.user, organizatoin=self.org)
        return self._mb

    def on_start(self):
        self.pending = {}
        self.org = self.membership.org
        logger.info('configured organization: %i', self.org)
        self.auth_token = self.membership.auth_token
        self.user_id = six.text_type(uuid.uuid4())

        if not self.host:
            raise AssertionError(
                '--host must be appended to determine the websocket location.'
            )

        # always use wss except if we are connecting to http.
        prot = 'wss' if urlparse(self.host).scheme != 'http' else 'ws'
        ws_location = '{}://{}/ws'.format(
            prot, self.host.replace('http://', '').replace('https://', '')
        )

        url = "{}?auth_token={}".format(ws_location, self.auth_token)
        logger.debug('connecting to %s' % url)
        try:
             # if necessary add  sslopt={"cert_reqs": ssl.CERT_NONE} to connect
            self.ws = ws = create_connection(
                url, origin=self.host, sslopt={"cert_reqs": ssl.CERT_REQUIRED}
            )
        except Exception as ex:
            error = 'cannot connect to websocket at %s. error was: %s' % (url, ex)
            logger.error(error)
            raise RuntimeError(error)
        if not ws:
            raise RuntimeError('cannot create websocket connection %s' % url)
        self.awaiting_login_confirmation = True

        def _receive():
            def reconnect():
                logger.info('reconnecting...')
                self.on_start()
                logger.info('reconnected.')

            while True:
                end_at = time()
                try:
                    res = ws.recv()
                except WebSocketConnectionClosedException as ex:
                    websocket_dropped.fire(ex=ex)
                    logger.exception(
                        'could not receive data from websocket, '
                        'reconnecting... exception was: %s', ex
                    )
                    request_failure.fire(
                        request_type='WEBSOCKET_DROP',
                        name='websocket dropped',
                        response_time=end_at,
                        exception=ex,
                    )
                    reconnect()
                    return

                try:
                    data = json.loads(res)
                except JSONDecodeError:
                    emsg = 'unable to decode response: %s' % res
                    logger.error(emsg)
                    request_failure.fire(
                        request_type='MALEFORMED RESPONSE',
                        name='Maleformed Response',
                        response_time=time(),
                        exception=RuntimeError(emsg),
                    )

                """
                response type cheat sheet for the hero extending this logic :).
                Message     Type    ID  Direction   Category
                WELCOME     0    Server-to-client    Auxiliary
                PREFIX      1    Client-to-server    Auxiliary
                CALL        2    Client-to-server    RPC
                CALLRESULT  3    Server-to-client    RPC
                CALLERROR   4    Server-to-client    RPC
                SUBSCRIBE   5    Client-to-server    PubSub
                UNSUBSCRIBE 6    Client-to-server    PubSub
                PUBLISH     7    Client-to-server    PubSub
                EVENT       8    Server-to-client    PubSub
                """

                call_id = data[1]
                response_type = data[0]
                try:
                    start_at = self.pending[call_id]['start_at']
                    name = self.pending[call_id]['name']
                    # clean up this dict otherwise it gets huuuge
                    del self.pending[call_id]
                except KeyError:
                    start_at = end_at
                    name = ''
                    # server to client has never a start time
                    if not self.awaiting_login_confirmation and response_type in (1, 2, 5, 6, 7):
                        logger.error(
                            'call %s has no start time or name', call_id
                        )
                    else:
                        self.awaiting_login_confirmation = False

                response_time = int((end_at - start_at) * 1000)
                event_type = 'WebSocket Recv'

                if response_type == 4:
                    """
                    last entry of wamp error response is either errorDetails on
                    pos 4 or errorDesc on pos 3 so no matter last entry it is.
                    """
                    error_msg = data.pop()
                    logger.error('CALLERROR: %s (%s)', data.pop(), error_msg)
                    request_failure.fire(
                        request_type=event_type,
                        name=name,
                        response_time=response_time,
                        exception=RuntimeError(error_msg),
                    )
                elif response_type == 8:
                    """
                    event
                    """
                    request_success.fire(
                        request_type='Event Revc',
                        name=str(data[1].replace("http://domain/", "")),
                        response_time=0,
                        response_length=len(res),
                    )
                else:
                    """
                    currently every other response type means success.
                    deciding what response type is requested might happen
                    via the `pending` variable on request time.
                    that could come in handy sometimes I guess.
                    """
                    request_success.fire(
                        request_type=event_type,
                        name=name,
                        response_time=response_time,
                        response_length=len(res),
                    )

        def _ping():
            while True:
                gevent.sleep(10)
                cid = generate_call_id()
                body = '[2,"{}","http://domain/ping"]'.format(cid)
                self._send_ws(cid, 'ping', body)

        gevent.spawn(_receive)
        gevent.spawn(_ping)

    def _send_ws(self, cid, name, body):
        """
        does the actual ws sending
        """
        start_at = time()
        self.pending[cid] = {
            'start_at': start_at,
            'name': name
        }
        logger.info('sending request with name %s for payload %s', name, body)
        logger.debug(body)
        try:
            self.ws.send(body)
        except WebSocketConnectionClosedException as ex:
            websocket_dropped.fire(ex=ex)
            request_failure.fire(
                request_type='WEBSOCKET_DROP',
                name='websocket dropped',
                response_time=start_at,
                exception=ex,
            )
            self.on_start()
            return

        request_success.fire(
            request_type='WebSocket Sent',
            name=name,
            # this basically is always 0ish as its just the delay of the
            # sending...but one might see connection delays. also we need
            # this count as to match whether all requests got responses
            response_time=int((time() - start_at) * 1000),
            response_length=len(body),
        )

    def send_ws(self, ns, action, *args, **kwargs):
        call_id = generate_call_id()
        positional = ','.join([str(a) for a in args]) if args else ''
        keyword = str(json.dumps(kwargs)) if kwargs else None
        payload = positional
        if kwargs:
            payload += ',' + keyword

        body = '[2,"{call_id}","http://domain/{ns}/{action}",{payload}]'.format(
            call_id=call_id, ns=ns, action=action, payload=payload
        )
        # this is t place for finding slow requests if the have special params.
        # just make the name more unique.

        name = '{}.{}.{}'.format('ws', ns, action)

        if 'description' in kwargs:
            name += f" ({kwargs.pop('description')})"

        self._send_ws(call_id, name, body)

    def send_lp(self, ns, action, *args, **kwargs):
        args += (kwargs,)
        name = '{}.{}.{}'.format('lp', ns, action)

        if 'description' in kwargs:
            name += f" ({kwargs.pop('description')})"

        params = {
            "args": args,
            "action": action,
            "ns": ns,
            "debug": {'name': name}
        }

        with self.client.post(
                "/lp/rpc/", data=json.dumps(params), name=name,
                catch_response=True,
                headers={
                    "Authorization": "Token {}".format(self.auth_token),
                    "Content-Type": "application/json; charset=utf-8",
                },
                verify=False,
        ) as response:
            logger.debug('get_users response: %i' % response.status_code)
            if response.status_code < 200 or response.status_code > 299:
                response.failure(response.status_code)

    def send(self, ns, action, *args, **kwargs):
        # default should be WS, long polling is disabled for now as it does
        # not reflect the true performance (events and subscriptions on ws).
        if WS:
            self.send_ws(ns, action, *args, **kwargs)
        else:
            self.send_lp(ns, action, *args, **kwargs)

    def on_quit(self):
        # wait for the last responses to finish.
        i = 0
        max_wait_sec = 5
        while len(self.pending) and i < max_wait_sec:
            logger.info(
                'waiting to shutdown websocket, %i responses missing.',
                len(self.pending)
            )
            gevent.sleep(1)
            i += 1
        if i == max_wait_sec:
            logger.error(
                'waited for %i seconds but there are still %i requests '
                'missing. killing websocket.', i, len(self.pending)
            )
        self.ws.close()
        self.logger.info('websocket closed.')



from cement import Controller, ex
from cement.utils import fs
import json
from benchgrape.core.db import TestDataMapper
from benchgrape.core.utils import map_benchmark_data


class TestData(Controller):
    class Meta:
        label = 'test-data'
        stacked_type = 'nested'
        stacked_on = 'base'

    @ex(help='Load Test Data from JSON-Config File generated by '
             'chatgrape\'s benchmark_setup command.',
        arguments=[
            (['config_file'], {
                'help': 'Config File exported by ChatGrape\'s '
                        'benchmark_setup command',
                'action': 'store'
            })
        ],
    )
    def load(self):
        fname = fs.abspath(self.app.pargs.config_file)
        self.app.log.info('loading config file from: %s' % fname)

        with open(fname, 'rb') as fhandle:
            config = json.loads(fhandle.read())

        organization, users, groups, private_conversations = map_benchmark_data(
            config
        )

        TestDataMapper(self.app.db, self.app.log).sync_db(
            organization, users, groups, private_conversations
        )

    @ex(help='Purge test data in the BenchGrape DB. '
             'This also reset the internal data structure.')
    def purge(self):
        TestDataMapper(self.app.db, self.app.log).drop_db()



import logging
import os
import six
import uuid
from locust import TaskSet

from benchgrape.core.statement import login

logger = logging.getLogger(__name__)


class LongPollingTaskSet(TaskSet):
    def on_start(self):
        self.org = int(os.getenv('ORG_ID', 1))
        logger.info('configured organization: %i', self.org)
        self.auth_token = login(self.client)
        self.user_id = six.text_type(uuid.uuid4())

        if not self.parent.host:
            raise AssertionError(
                '--host must be appended to determine the websocket location.'
            )




def map_benchmark_data(raw_data):
    """"
    map benchmark data do internal format -> currently just a proxy where
    we could interfere
    :param raw_data: ``dict``
    :return: ``dict``
    """
    organization = raw_data['organization']
    organization['url'] = raw_data['url']

    users = []
    for id, user_data in raw_data['users'].items():
        meta = user_data
        meta['id'] = id
        users.append(meta)

    groups = []
    for id, group_data in raw_data['groups'].items():
        meta = group_data
        meta['id'] = id
        groups.append(meta)

    pcs = []
    for id, pc_data in raw_data['private_conversations'].items():
        meta = pc_data
        meta['id'] = id
        pcs.append(meta)

    return organization, users, groups, pcs

    # import subprocess, time, os, sys
    #
    # cmd = "rsync.exe -vaz souce/ dest/"
    #
    # p = subprocess.Popen(cmd,
    #                      shell=True,
    #                      bufsize=64,
    #                      stdin=subprocess.PIPE,
    #                      stderr=subprocess.PIPE,
    #                      stdout=subprocess.PIPE)
    #
    # for line in p.stdout:
    #     print(">>> " + str(line.rstrip()))
    #     p.stdout.flush()
    #

#
# seek = 0
# line = ""
# extra_run = 5000  ## rsync.exe send returncode before terminated. Extra bytes.
# while True:
#     self.std.seek(seek)
#     byte = str(self.std.read(1), "windows-1252")
#     if byte == "\n" or byte == "\r":
#         print(line)
#         line = ""
#     else:
#         line += byte
#     seek += 1



from cement import Controller, ex
import subprocess


class Benchmark(Controller):
    class Meta:
        label = 'benchmark'
        stacked_type = 'nested'
        stacked_on = 'base'

    @ex(help='start the locust suit.')
    def start(self):
        subprocess.run([
            "locust",
            "--locustfile=./benchgrape/locust/locustfiles/stability.py",
            "--host=%s" % "https://staging.chatgrape.com/"
        ])



import logging
from locust import HttpLocust, task, seq_task
import random
import string

from benchgrape.core.statement import PreparedStatementsMixin
from benchgrape.core.wamp import WampTaskSet, PubSubMixin

logger = logging.getLogger(__name__)


DEFAULT_MIN_WAIT = 10
DEFAULT_MAX_WAIT = 3600 * 1000


def get_random_section():
    return random.choice(string.ascii_letters + '#')


def get_random_message():
    word = ["mind", "woebegone", "answer", "grip", "remain", "annoying", "green", "fowl", "macho", "alert", "regular", "steam", "fast", "lumber", "neighborly", "scintillating", "shade", "ragged", "majestic", "unit", "pig", "flat", "accidental", "library", "disturbed", "cellar", "confess", "actor", "successful", "godly", "sick", "skin", "blood", "round", "vase", "appreciate", "gun", "whole", "competition", "obey", "misty", "gabby", "cynical", "blush", "stem", "striped", "groan", "knee", "dead", "faithful", "ugliest", "fill", "damaging", "jail", "cross", "precede", "cute", "experience", "careful", "adhesive", "type", "boring", "towering", "meeting", "cable", "anxious", "approval", "grease", "plausible", "pretty", "tease", "copper", "material", "fat", "move", "pump", "whirl", "cow", "sweater", "grade", "cherries", "nail", "spy", "belong", "nine", "weight", "opposite", "follow", "water", "jumpy", "nutty", "crow", "exciting", "rustic", "school", "chalk", "few", "mate", "country", "team"]
    return " ".join(random.sample(word, random.randint(1, 25))) + "."


class PhoneBookTaskSet(WampTaskSet, PreparedStatementsMixin):
    # user behaviour, how much time between the actions, we need kibana to
    # monitor and fine tune this numbers to be realistic.
    # i'd like to add kibana to our staging and measure how activity looks like
    # in terms of API calls (which, how much time in between, how many at most
    # for how many online users, etc...)
    min_wait = 10
    max_wait = 3600 * 1000

    # users
    @task()
    def get_users_no_query_membership_true(self):
        self.users__get_users(True)

    @task()
    def get_users_no_query_membership_false(self):
        self.users__get_users(False)

    @task()
    def get_users_no_query_membership_none(self):
        self.users__get_users(None)

    @task()
    def get_users_query_membership_true(self):
        self.users__get_users(True, 'ad')

    @task()
    def get_users_query_membership_false(self):
        self.users__get_users(False, 'ad')

    @task()
    def get_users_query_membership_none(self):
        self.users__get_users(None, 'ad')

    @task()
    def get_users_sections(self):
        self.users__get_users(None, {'section': get_random_section()})

    @task()
    def get_rooms_no_query_membership_true(self):
        self.rooms__get_rooms(True)

    @task()
    def get_rooms_no_query_membership_false(self):
        self.rooms__get_rooms(False)

    @task()
    def get_rooms_no_query_membership_none(self):
        self.rooms__get_rooms()

    @task()
    def get_rooms_query_membership_true(self):
        self.rooms__get_rooms(True, 'ad')

    @task()
    def get_rooms_query_membership_false(self):
        self.rooms__get_rooms(False, 'ad')

    @task()
    def get_rooms_query_membership_none(self):
        self.rooms__get_rooms(None, 'ad')


class HistoryTaskSet(PubSubMixin, WampTaskSet, PreparedStatementsMixin):
    """
    simulate chat behaviour with posting, updating, typing, channel switching.
    """
    min_wait = 5 * 1000
    max_wait = 3600 * 1000

    @task(10)
    def post_message(self):
        self.channels__post(self.membership.get_random_channel(), get_random_message())

    @task(10)
    def get_overview(self):
        self.channels__get_overview()

    @task(30)
    def get_history_for_channel(self):
        self.channels__get_history(self.membership.get_random_channel())

    @task(30)
    def typing(self):
        self.channels__set_typing(
            self.membership.get_random_channel(), bool(random.randint(0, 1))
        )


class ReconnectTaskSet(PubSubMixin, WampTaskSet, PreparedStatementsMixin):
    """
    simulate a connection outage and reconnecting clients.
    """
    min_wait = 5 * 1000
    max_wait = 10 * 1000

    @seq_task(1)
    def disconnect(self):
        self.utils__disconnect_stack()

    @seq_task(2)
    def connect(self):
        self.utils__connect_stack()


class StandardActivityTaskSet(WampTaskSet, PreparedStatementsMixin):
    """
    container of subsets and their weights to simulate a normal activity in
    an organization. use this to benchmark server requirements for customers.
    """
    tasks = {
        #PhoneBookTaskSet: 5,
        ReconnectTaskSet: 10,
        #HistoryTaskSet: 80,
    }

    def on_start(self):
        super(StandardActivityTaskSet, self).on_start()
        # generate the load the client would do as well
        self.utils__connect_stack()


class IdleActivityTaskSet(WampTaskSet, PreparedStatementsMixin):
    """
    just establish the websocket, nothing more. see how many websockets we
    can handle in idle. - ping is still sent.
    """
    tasks = {}

    @task()
    def idle(self):
        pass

#
# class PhoneBookActivity(HttpLocust):
#     task_set = PhoneBookTaskSet
#
#
# class ChatActivity(HttpLocust):
#     task_set = HistoryTaskSet


class StandardActivity(HttpLocust):
    task_set = StandardActivityTaskSet



import logging
from locust import HttpLocust, task, seq_task

from benchgrape.core.statement import PreparedStatementsMixin
from benchgrape.core.wamp import WampTaskSet, PubSubMixin

logger = logging.getLogger(__name__)


DEFAULT_MIN_WAIT = 1000 * 60 * 5  # 5 min
DEFAULT_MAX_WAIT = 1000 * 60 * 15  # 15 min


class StabilityTaskSet(WampTaskSet, PreparedStatementsMixin):
    """
    stays connected and every 5-15 min sets a get_history call
    """
    tasks = {}

    @task()
    def idle(self):
        pass


class StabilityActivity(HttpLocust):
    task_set = StabilityTaskSet




from setuptools import setup, find_packages
from benchgrape.core.version import get_version

VERSION = get_version()

f = open('README.md', 'r')
LONG_DESCRIPTION = f.read()
f.close()

setup(
    name='benchgrape',
    version=VERSION,
    description='This is how you\'d Bench a Grape!',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Riccardo Salzer',
    author_email='beschwerden@salzamt.xyz',
    url='https://gitlab.chatgrape.com/salzamt/benchgrape',
    license='MIT',
    packages=find_packages(exclude=['ez_setup', 'tests*']),
    package_data={'benchgrape': ['templates/*']},
    install_requires=[
        'cement==3.0.2', 'locustio', 'websocket-client', 'jinja2', 'pyyaml',
        'colorlog',
    ],
    include_package_data=True,
    entry_points="""
        [console_scripts]
        benchgrape = benchgrape.main:main
    """,
)



from cement import Controller, ex
from locust.clients import HttpSession
from cement import fs
from benchgrape.core.db import TestDataMapper, RuntimeConfigMapper
from benchgrape.core.statement import login
import os
import subprocess


class Stability(Controller):
    class Meta:
        label = 'websocket'
        stacked_type = 'nested'
        stacked_on = 'base'

    @ex(help='establishes n connections to the server and make a long-running '
             'stability test of the websocket. log everything that seems '
             'unusual. can help a lot when debugging connection issues. '
             'WARNING: THIS DROPS ALL EXISTING BENCHMARK USERS!',
        arguments=[
            (
                    ['--url'], {
                        'help': 'url, example: http://chatgrape.com/',
                        'action': 'store',
                        'required': True
                    }
            ),
            (
                    ['--username'], {
                        'help': 'email/username used for login',
                        'action': 'store',
                        'required': True
                    }
            ),
            (
                    ['--password'], {
                        'help': 'password used for login. hint: 2fa is not '
                                'supported by this program. make sure you '
                                'either disable 2fa temporarily, or, create a '
                                'guest user for this test.',
                        'action': 'store',
                        'required': True
                    }
            ),
            (
                    ['--port'], {
                        'help': 'port to connect to, defaults to 443',
                        'action': 'store',
                        'default': '443'
                    }
            ),
            (
                    ['-n', '--number'], {
                        'help': 'number of websockets to connect, default 10',
                        'action': 'store',
                        'default': '10'
                    }
            ),
            (
                    ['-t', '--duration'], {
                        'help': 'duration of the benchmark in seconds',
                        'action': 'store',
                        'default': '3600'
                    }
            ),
            (
                    ['--logpath'], {
                        'help': 'Path to for files. '
                                'If not set, log will go to stdout/stderr',
                        'action': 'store',
                        'default': './logs'
                    }
            ),
        ],
    )
    def test(self):
        n_ws = self.app.pargs.number
        url = self.app.pargs.url
        username = self.app.pargs.username
        password = self.app.pargs.password
        duration = self.app.pargs.duration
        port = self.app.pargs.port
        logpath = fs.abspath(self.app.pargs.logpath)

        self.app.log.debug('creating log directory and log files.')
        fs.ensure_parent_dir_exists(logpath)
        fs.ensure_dir_exists(logpath)
        debug_fname = fs.join(logpath, 'debug.log')
        stats_fname = fs.join(logpath, 'stats.log')
        open(debug_fname, 'a').close()
        open(stats_fname, 'a').close()

        self.app.log.info('establish %s websocket connections to server %s '
                          'with user %s. press crtl+c to stop the test.' %
                          (n_ws, url, username))

        runtime_config = RuntimeConfigMapper(self.app.db, self.app.log)
        # make sure we always get the same user from the user backend by
        # enabling recycling and only have one user.
        runtime_config.enable_user_recycliing()

        mapper = TestDataMapper(self.app.db, self.app.log)
        self.app.log.info(
            'logging in...'
        )
        token = login(HttpSession(base_url=url), username, password)
        if not token:
            raise RuntimeError(
                f"login at {url} failed. stopping, wont connect to websocket."
            )
        self.app.log.info(
            'login successful. dropping all existing benchmark users to have '
            'a unique user to test with.'
        )
        mapper.drop_db()
        mapper.init_db()
        # int(u['id']), u['username'], u['password'], u['token']
        mapper.sync_db(
            {
                'id': 0, 'url': url, 'name': '', 'subdomain': ''
             }, [{
                    'id': 0, 'username': username, 'password': password,
                    'token': token
                }], [], []
        )
        self.app.log.info(
            'successfully added the test user, starting locust. '
            'visit the URL in the browser which will be displayed below '
            'in a few seconds.'
        )

        locust_file = fs.abspath(fs.join(
            os.path.dirname(__file__), '../locust/locustfiles/stability.py'
        ))
        self.app.log.info(f'using locustfile {locust_file}')
        assert os.path.isfile(locust_file), \
            f"locust file {locust_file} does not exist."

        l_args = [
             f"--locustfile={locust_file}",
             f"--host={url}",
             f"--logfile={fs.abspath(stats_fname)}",
        ]

        if n_ws:
            l_args.append(f"--clients={n_ws}")
        if duration:
            l_args.append(f"--run-time={duration}")
        if port:
            l_args.append(f"--port={port}")
        self.app.log.info(f'running locust {" ".join(l_args)}')
        subprocess.run([
            "locust",
            *l_args
        ])
